@@ -1,6 +1,8 @@
 use eframe::{egui, epi};
 
 fn main() {
+    puffin::set_scopes_on(true); // Remember to call this, or puffin will be disabled!
+
     let app = ExampleApp::default();
     let options = Default::default();
     eframe::run_native(Box::new(app), options);
@@ -16,10 +18,6 @@ impl epi::App for ExampleApp {
         "puffin egui eframe"
     }
 
-    fn setup(&mut self, _ctx: &egui::CtxRef) {
-        puffin::set_scopes_on(true); // Remember to call this, or puffin will be disabled!
-    }
-
     fn update(&mut self, ctx: &egui::CtxRef, _frame: &mut epi::Frame<'_>) {
         puffin::profile_function!();
         puffin::GlobalProfiler::lock().new_frame(); // call once per frame!
@@ -27,12 +25,24 @@ impl epi::App for ExampleApp {
         puffin_egui::profiler_window(ctx);
 
         // Give us something to inspect:
+
+        std::thread::Builder::new()
+            .name("Other thread".to_owned())
+            .spawn(|| {
+                sleep_ms(5);
+            })
+            .unwrap();
+
         sleep_ms(14);
         if self.frame_counter % 7 == 0 {
             puffin::profile_scope!("Spike");
             std::thread::sleep(std::time::Duration::from_millis(10))
         }
 
+        for _ in 0..1000 {
+            puffin::profile_scope!("very thin");
+        }
+
         self.frame_counter += 1;
     }
 }
